@@ -1,13 +1,44 @@
 
+static char rcsid[] =
+	"$Id: hoster.c,v 1.15 1999/08/19 15:41:06 pvmsrc Exp $";
+
 /*
-*	hoster.c
-*
-*	Example hoster task to demonstrate and verify the protocol.
-*	This code was cut out of the pvmd default hoster in startup.c,
-*	and minimally modified to make it compile here.
-*
-*	24 Jan 1994  Manchek
-*/
+ *         PVM version 3.4:  Parallel Virtual Machine System
+ *               University of Tennessee, Knoxville TN.
+ *           Oak Ridge National Laboratory, Oak Ridge TN.
+ *                   Emory University, Atlanta GA.
+ *      Authors:  J. J. Dongarra, G. E. Fagg, M. Fischer
+ *          G. A. Geist, J. A. Kohl, R. J. Manchek, P. Mucci,
+ *         P. M. Papadopoulos, S. L. Scott, and V. S. Sunderam
+ *                   (C) 1997 All Rights Reserved
+ *
+ *                              NOTICE
+ *
+ * Permission to use, copy, modify, and distribute this software and
+ * its documentation for any purpose and without fee is hereby granted
+ * provided that the above copyright notice appear in all copies and
+ * that both the copyright notice and this permission notice appear in
+ * supporting documentation.
+ *
+ * Neither the Institutions (Emory University, Oak Ridge National
+ * Laboratory, and University of Tennessee) nor the Authors make any
+ * representations about the suitability of this software for any
+ * purpose.  This software is provided ``as is'' without express or
+ * implied warranty.
+ *
+ * PVM version 3 was funded in part by the U.S. Department of Energy,
+ * the National Science Foundation and the State of Tennessee.
+ */
+
+/*
+ * hoster.c
+ *
+ * Example hoster task to demonstrate and verify the protocol.
+ * This code was cut out of the pvmd default hoster in startup.c,
+ * and minimally modified to make it compile here.
+ *
+ * 24 Jan 1994  Manchek
+ */
 
 #ifdef HASSTDLIB
 #include <stdlib.h>
@@ -15,8 +46,13 @@
 #include <stdio.h>
 #include <sys/types.h>
 #include <errno.h>
+#ifndef WIN32
 #include <pwd.h>
 #include <netdb.h>
+#endif
+#ifdef NEEDSSELECTH
+#include <sys/select.h>
+#endif
 #ifdef	SYSVSTR
 #include <string.h>
 #define	CINDEX(s,c)	strchr(s,c)
@@ -25,18 +61,22 @@
 #define	CINDEX(s,c)	index(s,c)
 #endif
 #include <pvm3.h>
-#include <pvmsdpro.h>
+#include <pvmproto.h>
+#include "bfunc.h"
+
+#ifdef WIN32
+#include "pvmwin.h"
+#endif
 
 #ifndef	RSHCOMMAND
 #define	RSHCOMMAND	"/usr/ucb/rsh"
 #endif
 
 #ifndef	RSHTIMEOUT
-#define	RSHTIMEOUT	60
+#define	RSHTIMEOUT 100
 #endif
 
 /* if > 1, uses parallel startup strategy */
-
 #ifndef	RSHNPLL
 #define	RSHNPLL	5
 #endif
@@ -49,14 +89,18 @@
 #define	min(a,b)	((a)<(b)?(a):(b))
 #endif
 
+#ifdef WIN32
+#define close(fd)	_close(fd)
+#endif
+
 #define	TALLOC(n,t,g)	(t*)malloc((n)*sizeof(t))
 #define	FREE(p)	free((char *)p)
 #define	STRALLOC(s)			strcpy(TALLOC(strlen(s)+1,char,"str"),s)
 
 #define	LISTPUTAFTER(o,n,f,r) \
-	{ (n)->f=(o)->f; (n)->r=o; (o)->f->r=n; (o)->f=n; }
+	{ (n)->f=(o)->f; (n)->r=(o); (o)->f->r=(n); (o)->f=(n); }
 #define	LISTPUTBEFORE(o,n,f,r) \
-	{ (n)->r=(o)->r; (n)->f=o; (o)->r->f=n; (o)->r=n; }
+	{ (n)->r=(o)->r; (n)->f=(o); (o)->r->f=(n); (o)->r=(n); }
 #define	LISTDELETE(e,f,r) \
 	{ (e)->f->r=(e)->r; (e)->r->f=(e)->f; (e)->r=(e)->f=0; }
 
@@ -98,18 +142,62 @@ struct hst {
 #define	HST_PASSWORD	1		/* ask for a password */
 #define	HST_MANUAL		2		/* do manual startup */
 	char *h_cmd;
+	char *h_wincmd;				/* alternate WIN32 default pvmd cmd */
 	char *h_result;
 };
 
+struct slot {
+	struct slot *s_link, *s_rlink;		/* free/active list */
+	struct hst *s_hst;					/* host table entry */
+	struct timeval s_bail;				/* timeout time */
+	int s_rfd, s_wfd, s_efd;			/* slave stdin/out/err */
+	char s_buf[256];					/* stdout from host */
+	char s_ebuf[256];					/* stderr from host */
+	int s_len;							/* length of s_buf */
+	int s_elen;							/* length of s_ebuf */
+};
 
+
+/***************
+ **  Globals  **
+ **           **
+ ***************/
+
+#ifndef HASERRORVARS
+extern int errno;
+#endif
+
+/* 0 turns off -- set to 1 to activate error messages in hoster */
 int debugmask = 0;
+
+#ifndef WIN32
 char *username = 0;
+#else
+extern char* username;
+extern int rsh( char *hostname, char *username, char *command,
+	char *retresult );
+#endif
+
+char *getenv();
+
+/***************
+ **  Private  **
+ **           **
+ ***************/
+
+static struct slot slots[RSHNPLL+2];	/* state var/ctx each slot */
+static struct slot *slfree = 0;			/* free list of slots */
 
 
+int
 main(argc, argv)
 	int argc;
 	char **argv;
 {
+#ifdef WIN32 
+	static int init = 0;
+	static int needed = 1;
+#else
 	struct passwd *pe;
 	int uid;
 
@@ -122,21 +210,57 @@ main(argc, argv)
 	else
 		fprintf(stderr, "main() can't getpwuid\n");
 	endpwent();
+#endif
 
 	pvm_mytid();
 	pvm_reg_hoster();
 	pvm_setopt(PvmResvTids, 1);
 
+#ifndef WIN32
 	printf("*** PVM pvmd starter ***\n");
+#endif
 
 	while (pvm_recv(-1, SM_STHOST) > 0) {
+
+#ifdef WIN32
+		AllocConsole();
+		
+		GetStdHandle(STD_OUTPUT_HANDLE);
+		if (init)  
+			printf(
+				"***need more ?                                 ***\n");
+#endif
+
 		hoster();
-		printf("***\n");
+
+#ifdef WIN32
+		if (!init) {
+			printf(
+				"*** PVM %s pvmd starter                                     ***\n",
+				PVM_VER);
+			printf(
+				"*** This window will show what your 'add command' is doing. ***\n");
+			printf(
+				"*** Also, if required, this is where you type in passwords. ***\n");
+			printf(
+				"*** This window will display each time you add machines.    ***\n");
+			Sleep(15000);	
+			FreeConsole();
+			init=1;
+		}
+		Sleep(1500);
+		FreeConsole();	
+#endif 
 	}
+
 	pvm_exit();
 	exit(0);
-}
 
+#ifdef WIN32
+	return(0);
+#endif
+}
+ 
 
 /*	hoster()
 *
@@ -154,8 +278,10 @@ hoster()
 	char sopts[64];
 	char lognam[256];
 	char cmd[512];
+	char wincmd[512];
 	int fromtid;
 	int wid;
+	struct pvmminfo minfo;
 
 	/*
 	* unpack the startup message
@@ -163,7 +289,8 @@ hoster()
 
 	pvm_bufinfo(pvm_getrbuf(), (int *)0, (int *)0, &fromtid);
 	pvm_unpackf("%d", &num);
-	wid = pvm_getmwid(pvm_getrbuf());
+	pvm_getminfo(pvm_getrbuf(), &minfo);
+	wid = minfo.wid;
 	printf("hoster() %d to start, wait id %d\n", num, wid);
 	if (num > 0) {
 		hostlist = TALLOC(num, struct hst *, "xxx");
@@ -172,7 +299,8 @@ hoster()
 			hostlist[i] = hp;
 			hp->h_flag = 0;
 			hp->h_result = 0;
-			if (pvm_unpackf("%d %s %s %s", &hp->h_tid, sopts, lognam, cmd)) {
+			if (pvm_unpackf("%d %s %s %s %s",
+					&hp->h_tid, sopts, lognam, cmd, wincmd)) {
 				printf("hoster() bad message format\n");
 				pvm_exit();
 				exit(1);
@@ -180,6 +308,13 @@ hoster()
 			hp->h_sopts = STRALLOC(sopts);
 			hp->h_login = STRALLOC(lognam);
 			hp->h_cmd = STRALLOC(cmd);
+
+			/* Check for (possible) alternate WIN32 pvmd cmd */
+			if (!strcmp(wincmd,""))
+				hp->h_wincmd = 0;
+			else
+				hp->h_wincmd = STRALLOC(wincmd);
+
 			printf("%d. t%x %s so=\"%s\"\n", i,
 					hp->h_tid,
 					hp->h_login,
@@ -206,14 +341,8 @@ hoster()
 	* do it
 	*/
 
-#if	RSHNPLL > 1
 	pl_startup(num, hostlist);
 
-#else	/*RSHNPLL > 1*/
-	for (i = 0; i < num; i++)
-		slave_exec(hostlist[i]);
-#endif	/*RSHNPLL > 1*/
-
 	/*
 	* send results back to pvmd
 	*/
@@ -228,33 +357,15 @@ hoster()
 /*
 	printf("hoster() sending back host table\n");
 */
-	pvm_setmwid(pvm_getsbuf(), wid);
+	pvm_getminfo(pvm_getsbuf(), &minfo);
+	minfo.wid = wid;
+	pvm_setminfo(pvm_getsbuf(), &minfo);
 	pvm_send(fromtid, SM_STHOSTACK);
 
 	return 0;
 }
 
 
-#if RSHNPLL > 1
-
-/********************************************
-*  this is the new (parallel) startup code  *
-*                                           *
-********************************************/
-
-struct slot {
-	struct slot *s_link, *s_rlink;		/* free/active list */
-	struct hst *s_hst;					/* host table entry */
-	struct timeval s_bail;				/* timeout time */
-	int s_rfd, s_wfd, s_efd;			/* slave stdin/out/err */
-	char s_buf[256];					/* config reply line */
-	int s_len;							/* length of s_buf */
-};
-
-
-static struct slot slots[RSHNPLL+2];	/* state var/context for each slot */
-static struct slot *slfree = 0;			/* free list of slots */
-
 close_slot(sp)
 	struct slot *sp;
 {
@@ -280,12 +391,16 @@ pl_startup(num, hostlist)
 	struct slot *sp, *sp2;
 	struct timeval tnow;
 	struct timeval tout;
+#ifdef FDSETNOTSTRUCT
+	fd_set rfds;
+#else
 	struct fd_set rfds;
+#endif
 	int nfds;
 	int i;
 	int n;
 	char *p;
-	char ebuf[256];						/* for reading stderr */
+	char c;
 
 	/* init slot free list */
 
@@ -320,6 +435,7 @@ pl_startup(num, hostlist)
 			LISTDELETE(sp, s_link, s_rlink);
 			sp->s_hst = hp;
 			sp->s_len = 0;
+			sp->s_elen = 0;
 			if (debugmask) {
 				fprintf(stderr, "pl_startup() trying %s\n", hp->h_name);
 			}
@@ -327,7 +443,45 @@ pl_startup(num, hostlist)
 			if (hp->h_result) {
 				/* error or fully started (manual startup) */
 
-				LISTPUTBEFORE(slfree, sp, s_link, s_rlink);
+				if (debugmask) {
+					fprintf(stderr,
+					"pl_startup() got result \"%s\" for %s phase1()\n",
+						hp->h_result, hp->h_name);
+				}
+
+				if (!strcmp(hp->h_result,"PvmCantStart")
+						&& hp->h_wincmd) {
+					if (debugmask) {
+						fprintf(stderr,
+							"pl_startup() re-trying %s as WIN32\n",
+							hp->h_name);
+					}
+					hp->h_cmd = sp->s_hst->h_wincmd;
+					hp->h_wincmd = 0;
+					FREE(hp->h_result);
+					hp->h_result = 0;
+
+					phase1(sp);
+
+					if (hp->h_result) {
+						if (debugmask) {
+							fprintf(stderr,
+					"pl_startup() got result \"%s\" for %s phase1()\n",
+								hp->h_result, hp->h_name);
+						}
+					} else {
+						/* partially started */
+
+						LISTPUTBEFORE(slact, sp, s_link, s_rlink);
+						gettimeofday(&sp->s_bail, (struct timezone*)0);
+						tout.tv_sec = RSHTIMEOUT;
+						tout.tv_usec = 0;
+						TVXADDY(&sp->s_bail, &sp->s_bail, &tout);
+					}
+
+				} else {
+					LISTPUTBEFORE(slfree, sp, s_link, s_rlink);
+				}
 
 			} else {
 				/* partially started */
@@ -355,7 +509,8 @@ pl_startup(num, hostlist)
 		gettimeofday(&tnow, (struct timezone*)0);
 		for (sp = slact->s_link; sp != slact; sp = sp->s_link) {
 			if (TVXLTY(&sp->s_bail, &tnow)) {
-				fprintf(stderr, "pl_startup() %s timed out after %d secs\n",
+				fprintf(stderr,
+					"pl_startup() giving up on host %s after %d secs\n",
 						sp->s_hst->h_name, RSHTIMEOUT);
 				sp->s_hst->h_result = STRALLOC("PvmCantStart");
 				sp2 = sp->s_rlink;
@@ -387,21 +542,35 @@ pl_startup(num, hostlist)
 			TVCLEAR(&tout);
 		}
 		if (debugmask) {
-			fprintf(stderr, "pl_startup() select timeout is %d.%06d\n",
+			fprintf(stderr, "pl_startup() select timeout is %d.%06ld\n",
 					tout.tv_sec, tout.tv_usec);
 		}
-		if ((n = select(nfds, &rfds, (fd_set*)0, (fd_set*)0, &tout)) == -1) {
+		if ((n = select(nfds,
+#ifdef  FDSETISINT
+				(int *)&rfds, (int *)0, (int *)0,
+#else
+				(fd_set *)&rfds, (fd_set *)0, (fd_set *)0,
+#endif
+				&tout)) == -1) {
 			if (errno != EINTR) {
-				pvmlogperror("work() select");
+#ifdef WIN32
+				pvmlogperror("pl_startup() select");
+#else
+				perror("pl_startup() select");
+#endif
 				pvmbailout(0);
 			}
 		}
 		if (debugmask) {
-			(void)fprintf(stderr, "pl_startup() select returns %d\n", n);
+			fprintf(stderr, "pl_startup() select returns %d\n", n);
 		}
 		if (n < 1) {
 			if (n == -1 && errno != EINTR) {
+#ifdef WIN32
 				pvmlogperror("pl_startup() select");
+#else
+				perror("pl_startup() select");
+#endif
 				pvmbailout(0);	/* XXX this is too harsh */
 			}
 			continue;
@@ -414,56 +583,24 @@ pl_startup(num, hostlist)
 		for (sp = slact->s_link; sp != slact; sp = sp->s_link) {
 
 			/*
-			* stdout ready.  get complete line then scan config info from it.
-			*/
-			if (sp->s_rfd >= 0 && FD_ISSET(sp->s_rfd, &rfds)) {
-				n = read(sp->s_rfd, sp->s_buf + sp->s_len,
-						sizeof(sp->s_buf) - sp->s_len);
-				if (n > 0) {
-					sp->s_len += n;
-					if (sp->s_len >= sizeof(sp->s_buf)) {
-						fprintf(stderr, "pl_startup() pvmd@%s: big read\n",
-								sp->s_hst->h_name);
-						sp->s_hst->h_result = STRALLOC("PvmCantStart");
-					}
-					sp->s_buf[sp->s_len] = 0;
-					if (p = CINDEX(sp->s_buf + sp->s_len - n, '\n')) {
-						if (debugmask) {
-							fprintf(stderr, "pvmd@%s: %s",
-									sp->s_hst->h_name, sp->s_buf);
-						}
-						*p = 0;
-						sp->s_hst->h_result = STRALLOC(sp->s_buf);
-					}
-
-				} else {
-					if (n) {
-						fprintf(stderr, "pl_startup() pvmd@%s\n",
-								sp->s_hst->h_name);
-						perror("");
-					} else {
-						fprintf(stderr, "pl_startup() pvmd@%s: EOF\n",
-								sp->s_hst->h_name);
-					}
-					sp->s_hst->h_result = STRALLOC("PvmCantStart");
-				}
-				if (sp->s_hst->h_result) {
-					sp2 = sp->s_rlink;
-					close_slot(sp);
-					sp = sp2;
-					continue;
-				}
-			}
-
-			/*
-			* response on stderr.  log prefixed by remote's host name.
+			* stderr ready.  log output with remote's host name.
 			*/
 			if (sp->s_efd >= 0 && FD_ISSET(sp->s_efd, &rfds)) {
-				if ((n = read(sp->s_efd, ebuf, sizeof(ebuf)-1)) > 0) {
-					char *p = ebuf, *q, c;
+#ifdef WIN32
+				n = win32_read_socket(sp->s_efd,
+						sp->s_ebuf + sp->s_elen,
+						sizeof(sp->s_ebuf) - sp->s_elen - 1);
+#else
+				n = read(sp->s_efd, sp->s_ebuf + sp->s_elen,
+						sizeof(sp->s_ebuf) - sp->s_elen - 1);
+#endif
+				if (n > 0) {
+					sp->s_elen += n;
+					sp->s_ebuf[sp->s_elen] = 0;
+#ifdef WIN32
 
-					ebuf[n] = 0;
-					fprintf(stderr, "pvmd@%s: ", sp->s_hst->h_name);
+					p = sp->s_ebuf + sp->s_elen - n;
+					fprintf(stderr, "stderr@%s: ", sp->s_hst->h_name);
 					while (c = *p++ & 0x7f) {
 						if (isprint(c))
 							fputc(c, stderr);
@@ -474,18 +611,161 @@ pl_startup(num, hostlist)
 						}
 					}
 					fputc('\n', stderr);
-
+#else
+					while (p = CINDEX(sp->s_ebuf, '\n')) {
+						*p = 0;
+						fprintf(stderr, "stderr@%s: %s\n",
+								sp->s_hst->h_name, sp->s_ebuf);
+						p++;
+						sp->s_elen -= p - sp->s_ebuf;
+						if (sp->s_elen > 0) {
+							BCOPY(p, sp->s_ebuf, sp->s_elen);
+							sp->s_ebuf[sp->s_elen] = 0;
+						}
+					}
+					if (sp->s_elen == sizeof(sp->s_ebuf) - 1) {
+						fprintf(stderr, "stderr@%s: %s\n",
+								sp->s_hst->h_name, sp->s_ebuf);
+						sp->s_elen = 0;
+					}
+#endif
 				} else {
+#ifndef WIN32
+					if (sp->s_elen > 0) {
+						fprintf(stderr, "stderr@%s: %s\n",
+								sp->s_hst->h_name, sp->s_ebuf);
+						sp->s_elen = 0;
+					}
+#endif
 					(void)close(sp->s_efd);
 					sp->s_efd = -1;
 				}
 			}
+
+			/*
+			* stdout ready.
+			* look for a complete line starting with "ddpro".
+			*/
+			if (sp->s_rfd >= 0 && FD_ISSET(sp->s_rfd, &rfds)) {
+#ifdef WIN32
+				n = win32_read_socket(sp->s_rfd, sp->s_buf + sp->s_len,
+						sizeof(sp->s_buf) - sp->s_len - 1);
+#else
+				n = read(sp->s_rfd, sp->s_buf + sp->s_len,
+						sizeof(sp->s_buf) - sp->s_len - 1);
+#endif
+				if (n > 0) {
+					sp->s_len += n;
+#ifdef WIN32
+					if (sp->s_len >= sizeof(sp->s_buf)) {
+						fprintf(stderr,
+								"pl_startup() pvmd@%s: big read\n",
+								sp->s_hst->h_name);
+						sp->s_hst->h_result = STRALLOC("PvmCantStart");
+					}
+#endif
+					sp->s_buf[sp->s_len] = 0;
+#ifdef WIN32
+					if (p = CINDEX(sp->s_buf + sp->s_len - n, '\n')) {
+						if (debugmask) {
+							fprintf(stderr, "pvmd@%s: %s",
+									sp->s_hst->h_name, sp->s_buf);
+						}
+						*p = 0;
+						sp->s_hst->h_result = STRALLOC(sp->s_buf);
+					}
+#else
+					while (p = CINDEX(sp->s_buf, '\n')) {
+						*p = 0;
+						if (!strncmp(sp->s_buf, "ddpro", 5)) {
+							if (debugmask) {
+								fprintf(stderr, "stdout@%s: %s\n",
+										sp->s_hst->h_name, sp->s_buf);
+							}
+							sp->s_hst->h_result = STRALLOC(sp->s_buf);
+							break;
+
+						} else {
+							fprintf(stderr, "stdout@%s: %s\n",
+									sp->s_hst->h_name, sp->s_buf);
+							p++;
+							sp->s_len -= p - sp->s_buf;
+							if (sp->s_len > 0) {
+								BCOPY(p, sp->s_buf, sp->s_len);
+								sp->s_buf[sp->s_len] = 0;
+							}
+						}
+					}
+					if (sp->s_len == sizeof(sp->s_buf) - 1) {
+						fprintf(stderr, "stdout@%s: %s\n",
+								sp->s_hst->h_name, sp->s_buf);
+						sp->s_len = 0;
+					}
+#endif
+				} else {
+#ifndef WIN32
+					if (sp->s_len > 0) {
+						fprintf(stderr, "stdout@%s: %s\n",
+								sp->s_hst->h_name, sp->s_buf);
+						sp->s_len = 0;
+					}
+#endif
+					if (n) {
+						fprintf(stderr, "pl_startup() stdout@%s\n",
+								sp->s_hst->h_name);
+					} else {
+						fprintf(stderr, "pl_startup() stdout@%s: EOF\n",
+								sp->s_hst->h_name);
+					}
+#ifndef WIN32
+					if (sp->s_elen > 0) {
+						fprintf(stderr, "stderr@%s: %s\n",
+								sp->s_hst->h_name, sp->s_ebuf);
+						sp->s_elen = 0;
+					}
+#endif
+
+					/* before failing, check for alternate WIN32 cmd */
+					if (sp->s_hst->h_wincmd) {
+						if (debugmask) {
+							fprintf(stderr,
+								"pl_startup() re-trying %s as WIN32\n",
+								hp->h_name);
+						}
+						sp->s_hst->h_cmd = sp->s_hst->h_wincmd;
+						sp->s_hst->h_wincmd = 0;
+						FREE(sp->s_hst->h_result);
+						sp->s_hst->h_result = 0;
+
+						phase1(sp);
+
+						if (!(sp->s_hst->h_result)) {
+							/* partially started */
+							gettimeofday(&sp->s_bail,
+								(struct timezone *) 0);
+							tout.tv_sec = RSHTIMEOUT;
+							tout.tv_usec = 0;
+							TVXADDY(&sp->s_bail, &sp->s_bail, &tout);
+						}
+
+					} else {
+						sp->s_hst->h_result = STRALLOC("PvmCantStart");
+					}
+				}
+				if (sp->s_hst->h_result) {
+					sp2 = sp->s_rlink;
+					close_slot(sp);
+					sp = sp2;
+					continue;
+				}
+			}
 		}
 	}
 	return 0;
 }
 
 
+
 phase1(sp)
 	struct slot *sp;
 {
@@ -497,6 +777,12 @@ phase1(sp)
 	int pid = -1;			/* pid of rsh */
 	char *p;
 
+#ifdef WIN32
+	char do_cmd[128];
+	char *val=0;
+	FILE   *faked_stream;
+#endif
+
 #ifndef NOREXEC
 	struct servent *se;
 	static u_short execport = 0;
@@ -507,7 +793,9 @@ phase1(sp)
 			pvmbailout(0);
 		}
 		execport = se->s_port;
+#ifndef WIN32
 		endservent();
+#endif
 	}
 #endif
 
@@ -524,6 +812,12 @@ phase1(sp)
 		fprintf(stderr, "Login to \"%s\" and type:\n", hn);
 		fprintf(stderr, "%s\n", hp->h_cmd);
 
+		if (hp->h_wincmd) {
+			fprintf(stderr, "\nor, if that command fails, " );
+			fprintf(stderr, "for WIN32 hosts try:\n");
+			fprintf(stderr, "%s\n\n", hp->h_wincmd);
+		}
+
 	/* get version */
 
 		fprintf(stderr, "Type response: ");
@@ -559,23 +853,24 @@ phase1(sp)
 		if (socketpair(AF_UNIX, SOCK_STREAM, 0, wpfd) == -1
 		|| socketpair(AF_UNIX, SOCK_STREAM, 0, rpfd) == -1
 		|| socketpair(AF_UNIX, SOCK_STREAM, 0, epfd) == -1) {
-			pvmlogperror("phase1() socketpair");
+			perror("phase1() socketpair");
 			goto oops;
 		}
 #else
+#ifndef WIN32
 		if (pipe(wpfd) == -1 || pipe(rpfd) == -1 || pipe(epfd) == -1) {
-			pvmlogperror("phase1() pipe");
+			perror("phase1() pipe");
 			goto oops;
 		}
 #endif
-
+#endif
+#ifndef WIN32
 		if (debugmask) {
 			fprintf(stderr, "phase1() pipes: %d %d %d %d %d %d\n",
 					wpfd[0], wpfd[1], rpfd[0], rpfd[1], epfd[0], epfd[1]);
 		}
-
 		if ((pid = fork()) == -1) {
-			pvmlogperror("phase1() fork");
+			perror("phase1() fork");
 			pvmbailout(0);
 		}
 		if (!pid) {
@@ -584,41 +879,98 @@ phase1(sp)
 			(void)dup2(epfd[1], 2);
 			for (i = getdtablesize(); --i > 2; )
 				(void)close(i);
+#else
+		{
+#endif
 			ac = 0;
+
 			av[ac++] = RSHCOMMAND;
 			av[ac++] = hn;
 			if (hp->h_login) {
 				av[ac++] = "-l";
 				av[ac++] = hp->h_login;
 			}
+#ifdef WIN32
+			else {
+				hp->h_login=username;
+				av[ac++] = "-l";
+				av[ac++] = hp->h_login;
+			}
+#endif
 			av[ac++] = hp->h_cmd;
+			/* av[ac++] = "-u"; WIN32 */
+			/* av[ac++] = username; WIN32 */
 			av[ac++] = 0;
 			if (debugmask) {
 				for (ac = 0; av[ac]; ac++)
 					fprintf(stderr, "av[%d]=\"%s\" ", ac, av[ac]);
 				fputc('\n', stderr);
 			}
+#ifndef WIN32
 			execvp(av[0], av);
 			fputs("phase1() execvp failed\n", stderr);
 			fflush(stderr);
 			_exit(1);
+#else
+			
+			sp->s_hst->h_result = malloc (512 * sizeof(char));
+
+			if (!rsh(hn,(hp->h_login ? hp->h_login : username),
+					hp->h_cmd,sp->s_hst->h_result)) {
+				fprintf(stderr, "phase1() rsh failed for host %s\n", hn);
+				goto oops;
+			}
+/* old code using popen ... */
+
+/*
+			faked_stream=_popen(do_cmd,"r"); 
+			
+			if (faked_stream != NULL)
+			{
+				while (1) {
+					fgets(buf,100,faked_stream);
+					if (strstr(buf,"ddpro")) {
+						if (!val)
+							val = malloc(sizeof (buf) * sizeof(char));
+						strcpy(val,buf);
+						sp->s_hst->h_result=val;
+						break;
+					
+					}
+					if (strcmp(buf,"PvmDupHost")){
+						if (!val)
+							val = malloc(sizeof (buf) * sizeof(char));
+						strcpy(val,buf);
+						sp->s_hst->h_result=val;
+						break;
+					}
+				}
+			}		
+*/
+#endif
 		}
+#ifndef WIN32
 		(void)close(wpfd[0]);
 		(void)close(rpfd[1]);
 		(void)close(epfd[1]);
 		sp->s_wfd = wpfd[1];
 		sp->s_rfd = rpfd[0];
 		sp->s_efd = epfd[0];
-
+#endif
 	} else {		/* use rexec to start */
 
 #ifdef NOREXEC
+#ifdef WIN32
 		fprintf(stderr, "slconfg() sorry, no rexec()\n");
+#else
+		fprintf(stderr, "slconfg() hoster was built without rexec support, can't take passwords\n");
+#endif
 		goto oops;
 #else
 		if (debugmask) {
 			fprintf(stderr, "phase1() rexec \"%s\"\n", hp->h_cmd);
 		}
+#ifndef WIN32
 		if ((sp->s_wfd = sp->s_rfd = rexec(&hn, execport,
 				(hp->h_login ? hp->h_login : username),
 				(char*)0, hp->h_cmd, &sp->s_efd))
@@ -626,13 +978,22 @@ phase1(sp)
 			fprintf(stderr, "phase1() rexec failed for host %s\n", hn);
 			goto oops;
 		}
+#else
+		sp->s_hst->h_result = malloc (512 * sizeof(char));
+
+		if (!rexec(hn,(hp->h_login ? hp->h_login : username),
+					hp->h_cmd,sp->s_hst->h_result))
+		{
+			fprintf(stderr, "phase1() rexec failed for host %s\n", hn);
+			goto oops;
+		}
+#endif
 #endif
 	}
 	return 0;
 
 oops:
 	hp->h_result = STRALLOC("PvmCantStart");
-oops2:
 	if (sp->s_wfd != -1)
 		close(sp->s_wfd);
 	if (sp->s_rfd != -1)
@@ -643,5 +1004,3 @@ oops2:
 	return 1;
 }
 
-#endif	/*RSHNPLL > 1*/
-
